#!/usr/bin/env python
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
###############################################################################
# procS1GMT5SAR.py
#
# Project:  apd_insar 
# Purpose:  Process single interferogram using GMT5SAR
#          
# Author:   Tom Logan
#
###############################################################################
# Copyright (c) 2017, Alaska Satellite Facility
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Library General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
# 
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Library General Public License for more details.
# 
# You should have received a copy of the GNU Library General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.
###############################################################################

#####################
#
# Import all needed modules right away
#
#####################
import sys, os, re
import argparse
import requests
import shutil
import glob
import zipfile

try:
    from lxml import html
    from lxml import etree
    import lxml.etree as et
except ImportError:
    raise ImportError("ERROR: Can not load module lxml; Please install it.")

try:
    from osgeo import gdal
except ImportError:
    raise ImportError("ERROR: Can not load module GDAL; Please install it.")
    
if int(gdal.VersionInfo()) < 2010000:
    print "ERROR: Must use GDAL version 2.1 or greater"
    sys.exit(1)
    

def getPageContents(url):
    page = requests.get(url)
    tree = html.fromstring(page.content)
    l = tree.xpath('//a[@href]/text()')
    ret = []
    for item in l:
        if 'EOF' in item:
            ret.append(item)
    return ret

def findOrbFile(plat,tm,lst):
    d1 = 0
    best = ''
    for item in lst:
        item = item.replace(' ','')
        item1 = item
        this_plat=item[0:3]
        item=item.replace('T','')
        item=item.replace('V','')
        t = re.split('_',item)
        start = t[6]
        end = t[7].replace('.EOF','')
        if start < tm and end > tm and plat == this_plat:
            d = ((int(tm)-int(start))+(int(end)-int(tm)))/2
            if d>d1:
                best = item1.replace(' ','')
    return best

def getOrbFile(s1Granule):
    url1 = 'https://s1qc.asf.alaska.edu/aux_poeorb/'
    url2 = 'https://s1qc.asf.alaska.edu/aux_resorb/'
    t = re.split('_+',s1Granule)
    st = t[4].replace('T','')
    # Try url1
    url = url1
    files = getPageContents(url)
    plat = s1Granule[0:3]
    orb = findOrbFile(plat,st,files)
    if orb == '':
        url = url2
        files = getPageContents(url)
        orb = findOrbFile(plat,st,files)
    return url+orb,orb

def get_cc(myxml):
    lon_max = -180
    lon_min = 180
    lat_max = -90
    lat_min = 90
    root = etree.parse(myxml)
    for coord in root.iter('latitude'):
         lat_max = max(float(coord.text),lat_max)
         lat_min = min(float(coord.text),lat_min)
    for coord in root.iter('longitude'):
         lon_max = max(float(coord.text),lon_max)
         lon_min = min(float(coord.text),lon_min)
    lat_max = lat_max + 0.15;
    lat_min = lat_min - 0.15;
    lon_max = lon_max + 0.15;
    lon_min = lon_min - 0.15;
    return lat_max,lat_min,lon_max,lon_min

def getDemFile(infile):
    mydir = "%s/annotation" %  infile
    myxml = ""
    name = ""

    # Get corners from first and last swath
    name = "001.xml"
    for myfile in os.listdir(mydir):
        if name in myfile:
            myxml = "%s/annotation/%s" % (infile,myfile)
    (lat1,lat2,lon1,lon2) = get_cc(myxml)
    name = "003.xml"
    for myfile in os.listdir(mydir):
        if name in myfile:
            myxml = "%s/annotation/%s" % (infile,myfile)
    (lat3,lat4,lon3,lon4) = get_cc(myxml)

    lon_max = -180
    lon_min = 180
    lat_max = -90
    lat_min = 90

    lat_max = max(lat1,lat2,lat3,lat4)
    lat_min = min(lat1,lat2,lat3,lat4)
    lon_max = max(lon1,lon2,lon3,lon4)
    lon_min = min(lon1,lon2,lon3,lon4)

    cmd = "wget -Odem.tif \"http://opentopo.sdsc.edu/otr/getdem?demtype=SRTMGL1&west=%s&south=%s&east=%s&north=%s&outputFormat=GTiff\"" % (lon_min,lat_min,lon_max,lat_max)
    print cmd
    os.system(cmd)

def linkRaw(image,pol):
    for name in glob.glob("../%s/annotation/s1*-*-slc-%s*.xml" % (image,pol)):
        new = os.path.basename(name)
        os.symlink(name,new)
    for name in glob.glob("../%s/measurement/s1*-*-slc-%s*.tiff" % (image,pol)):
        new = os.path.basename(name)
        os.symlink(name,new)
	
def alignSwaths(swath,pol,mdate,sdate,master_orbit,slave_orbit):
    master_iw = glob.glob("s1*-iw%s-slc-%s-%s*.tiff" % (swath,pol,mdate))[0]
    master_iw = master_iw.replace('.tiff','')
    slave_iw = glob.glob("s1*-iw%s-slc-%s-%s*.tiff" % (swath,pol,sdate))[0]
    slave_iw = slave_iw.replace('.tiff','')
    cmd = "align_tops.csh %s %s %s %s dem.grd" % (master_iw,master_orbit,slave_iw,slave_orbit)
    os.system(cmd)

def mkdirs(swath,pol,demname):
    os.mkdir("%s_%s" % (swath,pol))
    os.chdir("%s_%s" % (swath,pol))
    os.symlink("../config.s1a.txt","config.s1a.txt")
    os.mkdir("raw")
    os.chdir("raw")
    for myfile in glob.glob("../../raw_%s/*%s*" % (pol,swath)):
        new = os.path.basename(myfile)
        os.symlink(myfile,new)
    os.chdir("..")
    os.mkdir("topo")
    os.chdir("topo")
    os.symlink("../../%s" % demname,"dem.grd")
    os.chdir("../..")

def process_igram(swath,pol,mdate,sdate):
    os.chdir("%s_%s" % (swath,pol))
    mfile = glob.glob("raw/S1A%s_*_%s.SLC" % (mdate,swath))[0]
    mfile = os.path.basename(mfile)
    mfile = mfile.replace(".SLC","")
    sfile = glob.glob("raw/S1A%s_*_%s.SLC" % (sdate,swath))[0]
    sfile = os.path.basename(sfile)
    sfile = sfile.replace(".SLC","")
    cmd = "p2p_S1A_TOPS.csh %s %s config.s1a.txt > log 2>&1" % (mfile,sfile)
    print "Running Interferogram: %s" % cmd
    os.system(cmd)
    os.chdir("..")

def process_pol(pol,demname,master,slave,mdate,sdate,master_orbit,slave_orbit):
    os.mkdir("raw_%s" % pol)
    shutil.copy(master_orbit,"raw_%s" % pol)
    shutil.copy(slave_orbit,"raw_%s" % pol)

    # Link in all of the files that we need
    os.chdir("raw_%s" % pol)
    linkRaw(master,pol)
    linkRaw(slave,pol)
    os.symlink("../%s" % demname,"dem.grd")  

    # Align the individual swaths
    alignSwaths(1,pol,mdate,sdate,master_orbit,slave_orbit)
    alignSwaths(2,pol,mdate,sdate,master_orbit,slave_orbit)
    alignSwaths(3,pol,mdate,sdate,master_orbit,slave_orbit)
    os.chdir("..")

    # Make swath directories and links
    mkdirs("F1",pol,demname)
    mkdirs("F2",pol,demname)
    mkdirs("F3",pol,demname)

    # Process the data into inteferograms
    process_igram("F1",pol,mdate,sdate)
    process_igram("F2",pol,mdate,sdate)
    process_igram("F3",pol,mdate,sdate)

def rename_kml(inKml, name, outKml):

    print('Changing name field in KML file ({0})'.format(inKml))
    ns = {"ns":"http://earth.google.com/kml/2.1"}
    parser = et.XMLParser(remove_blank_text=True)
    doc = et.parse(inKml, parser)
    for node in doc.xpath('//ns:Document',  namespaces=ns):
        node[0].text = name
    print('Writing output file ({0})'.format(outKml))
    with open(outKml, 'w') as outF:
        outF.write(et.tostring(doc, pretty_print=True, xml_declaration=True,encoding='UTF-8'))
    outF.close()

    return None

def getFiles(swath,pol,mdate,sdate):

    for mydir in glob.glob("../%s_%s/intf/2*/" % (swath,pol)):
        os.chdir(mydir)
	
        new = "%s_%s_%s_%s_phase.kml" % (mdate,sdate,swath,pol)
        rename_kml("phasefilt_mask_ll.kml","%s_%s_%s_%s_phase.png" % (mdate,sdate,swath,pol),new)

        if os.path.isfile("unwrap_mask_ll.png"):
            new = "%s_%s_%s_%s_unwrap.kml" % (mdate,sdate,swath,pol)
            rename_kml("unwrap_mask_ll.kml","%s_%s_%s_%s_unwrap.png" % (mdate,sdate,swath,pol),new)

        # Create kmz files
        zipname = "%s_%s_%s_color_phase.kmz" % (mdate,sdate,swath)
        with zipfile.ZipFile(zipname,'w') as myzip:
            myzip.write("%s_%s_%s_%s_phase.kml" % (mdate,sdate,swath,pol))
            myzip.write("phasefilt_mask_ll.png")
        shutil.copy(zipname,"../../../PRODUCT/")

        oldname = "%s_%s_%s_%s_unwrap.kml" % (mdate,sdate,swath,pol)
        zipname = "%s_%s_%s_unw_phase.kmz" % (mdate,sdate,swath)
        if os.path.isfile(oldname):
            with zipfile.ZipFile(zipname,'w') as myzip:
                myzip.write("%s_%s_%s_%s_unwrap.kml" % (mdate,sdate,swath,pol))
                myzip.write("unwrap_mask_ll.png")
            shutil.copy(zipname,"../../../PRODUCT/")

        # Translate .grd files into geotiffs and create browse PNGs
        oldname = "display_amp_ll.grd"
        newname = "%s_%s_%s_amp.tif" % (mdate,sdate,swath)
        gdal.Translate(newname,oldname,outputSRS="EPSG:4326",creationOptions = ['COMPRESS=LZW'])
        shutil.copy(newname,"../../../PRODUCT/")
    
        oldname = "corr_ll.grd"
        newname = "%s_%s_%s_corr.tif" % (mdate,sdate,swath)
        gdal.Translate(newname,oldname,outputSRS="EPSG:4326",creationOptions = ['COMPRESS=LZW'])
        shutil.copy(newname,"../../../PRODUCT/")
    
        newname = "%s_%s_%s_color_phase.png" % (mdate,sdate,swath)
        oldname = "phasefilt_mask_ll.png"
        gdal.Translate(newname,oldname,format="PNG",width=0,height=1024)
        shutil.copy(newname,"../../../PRODUCT/")

        oldname = "unwrap_mask_ll.grd"
        newname = "%s_%s_%s_unw_phase.tif" % (mdate,sdate,swath)
        if os.path.isfile(oldname):
            gdal.Translate(newname,oldname,outputSRS="EPSG:4326",creationOptions = ['COMPRESS=LZW'])
            shutil.copy(newname,"../../../PRODUCT/")
            oldname = "unwrap_mask_ll.png"
            pngname = "%s_%s_%s_unw_phase.png"  % (mdate,sdate,swath)
            gdal.Translate(pngname,oldname,format="PNG",width=0,height=1024)
            shutil.copy(pngname,"../../../PRODUCT/")

        os.chdir("../../../PRODUCT")



def gatherResults(master,slave,mdate,sdate):
    if not os.path.isdir("PRODUCT"):
        os.mkdir("PRODUCT")
    os.chdir("PRODUCT")
    
    getFiles("F1","vv",mdate,sdate)
    getFiles("F2","vv",mdate,sdate)
    getFiles("F3","vv",mdate,sdate)
    
#    if "SDV" in master and "SDV" in slave:
#        getFiles("F1","vh",mdate,sdate)
#        getFiles("F2","vh",mdate,sdate)
#        getFiles("F3","vh",mdate,sdate)
   

###############################################################################
###############################################################################

parser = argparse.ArgumentParser(description="Create interferograms using GMT5SAR")
parser.add_argument("master",help="Master S1 file")
parser.add_argument("slave",help="Slave S1 file")
parser.add_argument("-d", "--dem", help="User specified DEM file (in GMT format)")
args = parser.parse_args()

print "Found master file %s" % args.master
print "Found slave file %s" % args.slave
if args.dem is not None:
   print "Found user input DEM file %s" % args.dem

if "zip" in args.master:
    mdate = re.split('_',args.master)[5][0:8]
    print "Unzipping master file"
    zip_ref = zipfile.ZipFile(args.master, 'r')
    zip_ref.extractall(".")
    zip_ref.close()    
    master = glob.glob("*%s*.SAFE" % mdate)[0]
elif "SAFE" in args.master:
    master = args.master
    mdate = re.split('_',master)[5][0:8]
else:
    print "Input master file must be either .zip or .SAFE"

if "zip" in args.slave:
    sdate = re.split('_',args.slave)[5][0:8]
    print "Unzipping slave file"
    zip_ref = zipfile.ZipFile(args.slave, 'r')
    zip_ref.extractall(".")
    zip_ref.close()    
    slave = glob.glob("*%s*.SAFE" % sdate)[0]
elif "SAFE" in args.master:
    slave = args.slave
    sdate = re.split('_',slave)[5][0:8]
else:
    print "Input slave file must be either .zip or .SAFE"

# Copy in the config file if none is present
template_dir = os.path.dirname(os.path.realpath(__file__))
template = '%s/config.s1a.txt' % template_dir
if not os.path.isfile('config.s1a.txt'):
    shutil.copy(template,'config.s1a.txt')    

# Get the orbit files
print "Getting orbit files"
(orburl,master_orbit) = getOrbFile(master)
if not os.path.isfile(master_orbit):
    cmd = 'wget ' + orburl
    os.system(cmd)
else: 
    print "Found master orbit file %s" % master_orbit

(orburl,slave_orbit) = getOrbFile(slave)
if not os.path.isfile(slave_orbit):
    cmd = 'wget ' + orburl
    os.system(cmd)
else:
    print "Found slave orbit file %s" % slave_orbit

# Get a DEM file if we need to
if args.dem is None:
    print "Getting DEM file"
    getDemFile(master)
    gdal.Translate('dem.grd','dem.tif',format="GMT",xRes=0.000833333333333,yRes=0.000833333333333)
    demname = "dem.grd"
else:
    demname = args.dem

# Clean up from past runs if necessary
if os.path.isdir("raw_vv"):
    print "Cleaning old raw_vv directory"
    shutil.rmtree("raw_vv")
    
if os.path.isdir("F1_vv"):
    print "Cleaning old F1_vv directory"
    shutil.rmtree("F1_vv")
    
if os.path.isdir("F2_vv"):
    print "Cleaning old F2_vv directory"
    shutil.rmtree("F2_vv")
    
if os.path.isdir("F3_vv"):
    print "Cleaning old F3_vv directory"
    shutil.rmtree("F3_vv")
    
if os.path.isdir("raw_vh"):
    print "Cleaning old raw_vh directory"
    shutil.rmtree("raw_vh")
    
if os.path.isdir("F1_vh"):
    print "Cleaning old F1_vh directory"
    shutil.rmtree("F1_vh")
    
if os.path.isdir("F2_vh"):
    print "Cleaning old F2_vh directory"
    shutil.rmtree("F2_vh")
    
if os.path.isdir("F3_vh"):
    print "Cleaning old F3_vh directory"
    shutil.rmtree("F3_vh")

# Process main polarization
process_pol("vv",demname,master,slave,mdate,sdate,master_orbit,slave_orbit)

# Process cross-polarization if it exists
# if "SDV" in master and "SDV" in slave:
#    process_pol("vh",demname,master,slave,mdate,sdate,master_orbit,slave_orbit)

gatherResults(master,slave,mdate,sdate)

